#include <strstream.h>

#include "annulus.h"
#include "frame.h"
#include "util.h"

int rCompare(const void*, const void*);

Annulus::Annulus(Frame* p, const Vector& ctr, double r, 
	       const char* clr, const char* f, const char* t, 
	       unsigned short prop)
  : BaseCircle(p, ctr, r, clr, f, t, prop)
{
  annuli = 1;
  radii = new double[annuli];

  radii[0] = r;

  strcpy(type, "annulus");
  numHandle = 4 + annuli;
  handle = new Vector[numHandle];

  updateBBox();
}

Annulus::Annulus(Frame* p, const Vector& ctr, 
		 double inner, double outer, int num,
		 const char* clr, const char* f, const char* t, 
		 unsigned short prop)
  : BaseCircle(p, ctr, inner, clr, f, t, prop)
{
  annuli = num+1;
  radii = new double[annuli];

  for (int i=0; i<annuli; i++)
    radii[i] = i*(outer-inner)/num+inner;

  strcpy(type, "annulus");
  numHandle = 4 + annuli;
  handle = new Vector[numHandle];

  updateBBox();
}

Annulus::Annulus(Frame* p, const Vector& ctr, double* r, int an,
		 const char* clr, const char* f, const char* t, 
		 unsigned short prop)
  : BaseCircle(p, ctr, r[0], clr, f, t, prop)
{
  annuli = an;
  radii = new double[annuli];

  for (int i=0; i<annuli; i++)
    radii[i] = r[i];
  sortRadii();

  strcpy(type, "annulus");
  numHandle = 4 + annuli;
  handle = new Vector[numHandle];

  updateBBox();
}

void Annulus::updateBBox()
{
  // handles are in canvas coords
  // bound marker

  // we can't garantee that the radii have been sorted yet
  double max = 0;
  for(int i=0; i<annuli; i++)
    if (max<radii[i])
      max = radii[i];

  Vector c = center * parent->refToCanvas;
  Vector r = Vector(max,max);
  if (!(properties & FIXED))
    r *= parent->getZoom();
  bbox = BBox(c-r, c+r);

  // generate handles

  handle[0] = bbox.ll;
  handle[1] = bbox.lr();
  handle[2] = bbox.ur;
  handle[3] = bbox.ul();

  // the rest are radii

  {
    for (int i=0; i<annuli; i++) {
      Vector rr = Vector(radii[i],radii[i]);
      if (!(properties & FIXED))
	rr *= parent->getZoom();
      handle[i+4] = Vector(c[0]+rr[0], c[1]);
    }
  }

  // bound handles

  {
    for (int i=0; i<numHandle; i++)
      bbox.bound(handle[i]);
  }

  // make room for handles

  bbox.expand(3);

  // calculate overall bbox

  calcAllBBox();
}

void Annulus::edit(const Vector& v, int h)
{
  if (h<5) {
    for (int i=0; i<annuli; i++) {
      Vector r = (v-center).abs()/radii[annuli-1]*radii[i];
      radii[i] = r[0] > r[1] ? r[0] : r[1];
    }
  }
  else {
    Vector r = (v-center).abs();
    radii[h-5] = r[0] > r[1] ? r[0] : r[1];
  }

  updateBBox();
  doCallBacks(&editCB);
}

void Annulus::editEnd()
{
  sortRadii();

  updateBBox();
  doCallBacks(&editCB);
}

void Annulus::setRadii(double inner, double outer, int num)
{
  annuli = num+1;
  if (radii)
    delete [] radii;
  radii = new double[annuli];

  for (int i=0; i<annuli; i++)
    radii[i] = i*((outer-inner)/num)+inner;
  sortRadii();

  numHandle = 4 + annuli;
  delete [] handle;
  handle = new Vector[numHandle];

  updateBBox();
  doCallBacks(&editCB);
}

void Annulus::setRadii(const double* r, int num)
{
  annuli = num;
  if (radii)
    delete [] radii;
  radii = new double[annuli];

  for (int i=0; i<annuli; i++)
    radii[i] = r[i];
  sortRadii();

  numHandle = 4 + annuli;
  delete [] handle;
  handle = new Vector[numHandle];

  updateBBox();
  doCallBacks(&editCB);
}

int Annulus::addRadii(const Vector& v)
{
  // new radii array
  double* old = radii;
  radii = new double[annuli+1];

  // copy old values
  for (int i=0; i<annuli; i++)
    radii[i] = old[i];

  // delete old
  if (old)
    delete [] old;

  // new radii on end
  Vector r = (v-center).abs();
  radii[annuli] = r[0] > r[1] ? r[0] : r[1];
  annuli++;

  numHandle = 4 + annuli;
  delete [] handle;
  handle = new Vector[numHandle];

  updateBBox();
  //  doCallBacks(&editCB);

  // return number of radii
  return annuli;
}

void Annulus::deleteRadii(int h)
{
  if (annuli>1 && h<annuli) {
    // new radii array
    double* old = radii;
    radii = new double[annuli-1];

    // copy up to radii in question
    {
      for (int i=0; i<h; i++)
	radii[i] = old[i];
    }

    // copy remainder
    for (int i=h; i<annuli-1; i++)
      radii[i] = old[i+1];

    if (old)
      delete [] old;
    annuli--;

    numHandle = 4 + annuli;
    delete [] handle;
    handle = new Vector[numHandle];

    updateBBox();
    doCallBacks(&editCB);
  }
}

// list

void Annulus::list(ostream& str, int sys, int format, const char* delim)
{
  listCoordSystem(str, sys);
  str << ';';

  if (!(properties&INCLUDE))
    str << '-';

  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::PHYSICAL:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',';
      for (int i=0; i<annuli; i++) {
	str << parent->mapLenFromRef(radii[i], (Frame::CoordSystem)sys);
	if (i!=annuli-1)
	  str << ',';
      }
      str <<  ')';
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',';
	for (int i=0; i<annuli; i++) {
	  str << parent->mapLenFromRef(radii[i], 
				    (Frame::CoordSystem)sys,
				    Frame::ARCSEC) << '"';
	  if (i!=annuli-1)
	    str << ',';
	}
	str <<  ')';
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	str << type << '(' << ra << ',' << dec << ',';
	for (int i=0; i<annuli; i++) {
	  str << parent->mapLenFromRef(radii[i], 
				    (Frame::CoordSystem)sys, 
				    Frame::ARCSEC) << '"';
	  if (i!=annuli-1)
	    str << ',';
	}
	str <<  ')';
      }
      break;
    }
    break;
  }

  listProperties(str);
  str << delim;
}

void Annulus::listCiao(ostream& str, int sys, int format, const char* delim)
{
  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::PHYSICAL:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      for (int i=0; i<annuli-1; i++) {
	str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ','
	    << parent->mapLenFromRef(radii[i], (Frame::CoordSystem)sys) << ','
	    << parent->mapLenFromRef(radii[i+1], (Frame::CoordSystem)sys)
	    << ')';

	if (i!=annuli-2)
	  str << " & ";
      }
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << type << '(' << setprecision(8) << v[0] << "d," << v[1] << "d,";
	for (int i=0; i<annuli; i++) {
	  str << parent->mapLenFromRef(radii[i], 
				    (Frame::CoordSystem)sys,
				    Frame::ARCSEC) << '"';
	  if (i!=annuli-1)
	    str << ',';
	}
	str <<  ')';
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	str << type << '(' << ra << ',' << dec << ',';
	for (int i=0; i<annuli; i++) {
	  str << parent->mapLenFromRef(radii[i], 
				    (Frame::CoordSystem)sys, 
				    Frame::ARCSEC) << '"';
	  if (i!=annuli-1)
	    str << ',';
	}
	str <<  ')';
      }
      break;
    }
    break;
  }

  str << delim;
}

void Annulus::listFtools(ostream& str, int sys, int format, const char* delim)
{
  Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
  for (int i=0; i<annuli-1; i++) {
    str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ','
	<< parent->mapLenFromRef(radii[i], (Frame::CoordSystem)sys) << ','
	<< parent->mapLenFromRef(radii[i+1], (Frame::CoordSystem)sys) << ')';

    if (i!=annuli-2)
      str << " & ";
  }

  str << delim;
}

void Annulus::listPros(ostream& str, int sys, int format, const char* delim)
{
  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      str << type << ' ' << setprecision(8) 
	  << parent->map(center,Frame::REF,Frame::IMAGE);
      for (int i=0; i<annuli; i++)
	str << parent->mapLenFromRef(radii[i], Frame::IMAGE) << ' ';
    }
    break;

  case Frame::PHYSICAL:
    {
      str << "physical; ";
      str << type << ' ' << setprecision(8) 
	  << parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      for (int i=0; i<annuli; i++)
	str << parent->mapLenFromRef(radii[i], (Frame::CoordSystem)sys) << ' ';
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    listProsCoordSystem(str,sys);
    str << "; ";

    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << type << ' ' << setprecision(8) << v[0] << "d " << v[1] << "d ";
	for (int i=0; i<annuli; i++)
	  str << parent->mapLenFromRef(radii[i],
				    (Frame::CoordSystem)sys,
				    Frame::ARCSEC) << "\" ";
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	if (dec[0]=='+')
	  str << type << ' ' << ra << ' ' << dec+1 << ' ';
	else
	  str << type << ' ' << ra << ' ' << dec << ' ';

	for (int i=0; i<annuli; i++)
	  str << parent->mapLenFromRef(radii[i],
				    (Frame::CoordSystem)sys,
				    Frame::ARCSEC) << "\" ";
      }
      break;
    }
    break;
  }

  str << delim;
}

void Annulus::listSAOimage(ostream& str, const char* delim)
{
  // all coords are in image coords

  if (!(properties&INCLUDE))
    str << '-';

  Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
  str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',';

  for (int i=0; i<annuli; i++) {
    str << parent->mapLenFromRef(radii[i], Frame::IMAGE);
    if (i!=annuli-1)
      str << ',';
  }
  str <<  ')';

  str << delim;
}

void Annulus::sortRadii()
{
  qsort((void*)radii, annuli, sizeof(double), dCompare);
}

