#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>

#include <unistd.h>

#include <iostream.h>
#include <strstream.h>
#include <fstream.h>
#include <iomanip.h>

#include <Xlib.h>
#include <Xutil.h>

#include "frame.h"
#include "util.h"
#include "fitsimage.h"

#include "fvcontour.h"
#include "grid.h"

// utility member functions for cmds

void Frame::printIntegerTcl(int i)
{
  char buf[16];
  ostrstream str(buf,16);
  str << i << ends;
  Tcl_AppendResult(interp, buf, NULL);
}

void Frame::printDoubleTcl(double d, Precision p)
{
  char buf[64];
  ostrstream str(buf,64);

  switch (p) {
  case DEFAULT:
    str << d << ends;
    break;
  case FIXED:
    str << setiosflags(ios::fixed) << setw(9) << setprecision(3) << d << ends;
    break;
  case SCIENTIFIC:
    str << setiosflags(ios::scientific) << setprecision(8) << d << ends;
    break;
  case INTEGER:
    str << (int)d << ends;
    break;
  }

  Tcl_AppendResult(interp, buf, NULL);
}

void Frame::printVectorTcl(const Vector& v, Precision p)
{
  char buf[64];
  ostrstream str(buf,64);

  switch (p) {
  case DEFAULT:
    str << v << ends;
    break;
  case FIXED:
    str << setiosflags(ios::fixed) << setw(9) << setprecision(3) << v << ends;
    break;
  case SCIENTIFIC:
    str << setiosflags(ios::scientific) << setprecision(8) << v << ends;
    break;
  case INTEGER:
    {
      Vector z = v;
      str << z.round() << ends;
      break;
    }
  }

  Tcl_AppendResult(interp, buf, NULL);
}

void Frame::printCoordTcl(const Vector& v, CoordSystem in, 
			  CoordSystem out, Precision p)
{
  int valid;
  Vector r = map(v, in, out, &valid);
  if (valid)
    printVectorTcl(r, p);
}

void Frame::coordToTclArray(const Vector& v, CoordSystem in, CoordSystem out, 
			    Precision p, const char* var, const char* index)
{
  int valid;
  Vector r = map(v, in, out, &valid);

  doubleToTclArray(r[0], p, var, index, "x", valid);
  doubleToTclArray(r[1], p, var, index, "y", valid);
}

void Frame::doubleToTclArray(double d, Precision p, const char* var, 
			     const char* base, const char* mod, int valid)
{
  char index[64];
  strcpy(index,mod);
  strcat(index,base);

  if (valid) {
    char buf[64];
    ostrstream str(buf,64);

    switch (p) {
    case DEFAULT:
      str << d << ends;
      break;
    case FIXED:
      str << setiosflags(ios::fixed) << setw(9) << setprecision(3) 
	  << d << ends;
      break;
    case SCIENTIFIC:
      str << setiosflags(ios::scientific) << setprecision(8) << d << ends;
      break;
    case INTEGER:
      str << (int)d << ends;
      break;
    }

    Tcl_SetVar2(interp, (char*)var, index, buf, 0);
  }
  else
    Tcl_SetVar2(interp, (char*)var, index, "", 0);
}

// Frame Command Member Functions

void Frame::clearCmd()
{
  unloadFits();

  reset();

  updateClip();
  update(MATRIX);
}

void Frame::colorScaleCmd(ColorScaleType s)
{
  colorScaleType = s;
  updateColorScale();

  update(BASE);
}

void Frame::getColormapCmd()
{
  char buf[64];
  ostrstream str(buf,64);
  str << cmapID << ' ' << bias << ' ' << contrast << ' ' << invert << ends;

  Tcl_AppendResult(interp, buf, NULL);
}

void Frame::getColorScaleCmd()
{
  switch (colorScaleType) {
  case LINEARSCALE:
    Tcl_AppendResult(interp, "linear", NULL);
    break;
  case LOGSCALE:
    Tcl_AppendResult(interp, "log", NULL);
    break;
  case SQUAREDSCALE:
    Tcl_AppendResult(interp, "squared", NULL);
    break;
  case SQRTSCALE:
    Tcl_AppendResult(interp, "sqrt", NULL);
    break;
  case IISSCALE:
    Tcl_AppendResult(interp, "linear", NULL);
    break;
  }
}

void Frame::getCursorCmd(CoordSystem out, CoordFormat format)
{
  switch (out) {
  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    getWCSCoordCmd(cursor, USER, out, format);
    break;
  default:
    printVectorTcl(map(cursor, USER, out), DEFAULT);
  }
}

void Frame::getInfoCmd(char* var)
{
  if (fits) {
    Tcl_SetVar2(interp,var,"fileName",(char*)fits->getFileName(),0);
    Tcl_SetVar2(interp,var,"objectName",(char*)fits->getObjectName(),0);
    Tcl_SetVar2(interp,var,"min",(char*)fits->getMin(),0);
    Tcl_SetVar2(interp,var,"max",(char*)fits->getMax(),0);
    Tcl_SetVar2(interp,var,"low",(char*)fits->getLow(),0);
    Tcl_SetVar2(interp,var,"high",(char*)fits->getHigh(),0);
  }
  else {
    Tcl_SetVar2(interp,var,"objectName","",0);
    Tcl_SetVar2(interp,var,"fileName","",0);
    Tcl_SetVar2(interp,var,"min","",0);
    Tcl_SetVar2(interp,var,"max","",0);
    Tcl_SetVar2(interp,var,"low","",0);
    Tcl_SetVar2(interp,var,"high","",0);
  }

  Tcl_SetVar2(interp,var,"value","",0);
  Tcl_SetVar2(interp,var,"xImage","",0);
  Tcl_SetVar2(interp,var,"yImage","",0);
  Tcl_SetVar2(interp,var,"xPhysical","",0);
  Tcl_SetVar2(interp,var,"yPhysical","",0);
  Tcl_SetVar2(interp,var,"xAmplifier","",0);
  Tcl_SetVar2(interp,var,"yAmplifier","",0);
  Tcl_SetVar2(interp,var,"xDetector","",0);
  Tcl_SetVar2(interp,var,"yDetector","",0);

  Tcl_SetVar2(interp,var,"xWCS","",0);
  Tcl_SetVar2(interp,var,"yWCS","",0);
  Tcl_SetVar2(interp,var,"sysWCS","",0);
  Tcl_SetVar2(interp,var,"xUnitWCS","",0);
  Tcl_SetVar2(interp,var,"yUnitWCS","",0);
}

void Frame::getInfoCmd(const Vector& v, CoordSystem ref,
		       CoordSystem wcs, CoordFormat format, char* var)
{
  switch (ref) {
  case CANVAS:
    getInfo(v, ref, &FitsImage::getCanvasToData, &FitsImage::getCanvasToImage,
	    wcs, format, var);
    break;
  case PANNER:
    getInfo(v, ref, &FitsImage::getPannerToData, &FitsImage::getPannerToImage,
	    wcs, format, var);
    break;
  }
}

void Frame::getInfo(const Vector& v, CoordSystem ref,
		    Matrix& (FitsImage::*matrixToData)(),
		    Matrix& (FitsImage::*matrixToImage)(),
		    CoordSystem wcs, CoordFormat format, char* var)
{
  if (fitsCount>1)
    getInfoMosaic(v,ref,matrixToData,matrixToImage,wcs,format,var);
  else
    getInfoSingle(v,ref,matrixToData,matrixToImage,wcs,format,var);
}

void Frame::getInfoSingle(const Vector& v, CoordSystem ref,
			  Matrix& (FitsImage::*matrixToData)(),
			  Matrix& (FitsImage::*matrixToImage)(),
			  CoordSystem wcs, CoordFormat format, char* var)
{
  if (fits) {
    Vector img = v * (fits->*matrixToData)();
    Tcl_SetVar2(interp,var,"fileName",(char*)fits->getFileName(),0);
    Tcl_SetVar2(interp,var,"objectName",(char*)fits->getObjectName(),0);
    Tcl_SetVar2(interp,var,"min",(char*)fits->getMin(),0);
    Tcl_SetVar2(interp,var,"max",(char*)fits->getMax(),0);
    Tcl_SetVar2(interp,var,"low",(char*)fits->getLow(),0);
    Tcl_SetVar2(interp,var,"high",(char*)fits->getHigh(),0);

    if (img[0]>=0 && img[0]<fits->width() && img[1]>=0 && 
	img[1]<fits->height()) {
      Tcl_SetVar2(interp,var,"value",(char*)fits->getValue(img),0);

      coordToTclArray(v, ref, IMAGE, FIXED, var, "Image");
      coordToTclArray(v, ref, PHYSICAL, FIXED, var, "Physical");
      coordToTclArray(v, ref, AMPLIFIER, FIXED, var, "Amplifier");
      coordToTclArray(v, ref, DETECTOR, FIXED, var, "Detector");

      if (fits->hasWCS()) {
	char buf[64];
	fits->pix2wcs(v*(fits->*matrixToImage)(), wcs, format, buf, 64);

	int argc;
	char** argv;
	Tcl_SplitList(interp, buf, &argc, &argv);

	if (argv) {
	  if (argc > 0 && argv[0])
	    Tcl_SetVar2(interp,var,"xWCS",argv[0],0);
	  else
	    Tcl_SetVar2(interp,var,"xWCS","",0);

	  if (argc > 1 && argv[1])
	    Tcl_SetVar2(interp,var,"yWCS",argv[1],0);
	  else
	    Tcl_SetVar2(interp,var,"yWCS","",0);

	  if (argc > 2 && argv[2])
	    Tcl_SetVar2(interp,var,"sysWCS",argv[2],0);
	  else
	    Tcl_SetVar2(interp,var,"sysWCS","",0);

	  if (argc > 3 && argv[3])
	    Tcl_SetVar2(interp,var,"xUnitWCS",argv[3],0);
	  else
	    Tcl_SetVar2(interp,var,"xUnitWCS","",0);

	  if (argc > 4 && argv[4])
	    Tcl_SetVar2(interp,var,"yUnitWCS",argv[4],0);
	  else
	    Tcl_SetVar2(interp,var,"yUnitWCS","",0);

	  Tcl_Free((char*)argv);
	}
	else
	  goto noWCS;
      }
      else 
	goto noWCS;
    }
    else
      goto noImage;

    return;
  }

  // else, return blanks

  Tcl_SetVar2(interp,var,"objectName","",0);
  Tcl_SetVar2(interp,var,"fileName","",0);
  Tcl_SetVar2(interp,var,"min","",0);
  Tcl_SetVar2(interp,var,"max","",0);
  Tcl_SetVar2(interp,var,"low","",0);
  Tcl_SetVar2(interp,var,"high","",0);

noImage:
  Tcl_SetVar2(interp,var,"value","",0);
  Tcl_SetVar2(interp,var,"xImage","",0);
  Tcl_SetVar2(interp,var,"yImage","",0);
  Tcl_SetVar2(interp,var,"xPhysical","",0);
  Tcl_SetVar2(interp,var,"yPhysical","",0);
  Tcl_SetVar2(interp,var,"xAmplifier","",0);
  Tcl_SetVar2(interp,var,"yAmplifier","",0);
  Tcl_SetVar2(interp,var,"xDetector","",0);
  Tcl_SetVar2(interp,var,"yDetector","",0);

noWCS:
  Tcl_SetVar2(interp,var,"xWCS","",0);
  Tcl_SetVar2(interp,var,"yWCS","",0);
  Tcl_SetVar2(interp,var,"sysWCS","",0);
  Tcl_SetVar2(interp,var,"xUnitWCS","",0);
  Tcl_SetVar2(interp,var,"yUnitWCS","",0);
}

void Frame::getInfoMosaic(const Vector& v, CoordSystem ref,
			  Matrix& (FitsImage::*matrixToData)(),
			  Matrix& (FitsImage::*matrixToImage)(),
			  CoordSystem wcs, CoordFormat format, char* var)
{
  if (fits) {
    FitsImage* ptr = fits;
    while (ptr) {
      Vector img = v * (ptr->*matrixToData)();
      if (img[0]>=0 && img[0]<ptr->width() && 
	  img[1]>=0 && img[1]<ptr->height()) {
	Tcl_SetVar2(interp,var,"fileName",(char*)ptr->getFileName(),0);
	Tcl_SetVar2(interp,var,"objectName",(char*)ptr->getObjectName(),0);
	Tcl_SetVar2(interp,var,"min",(char*)ptr->getMin(),0);
	Tcl_SetVar2(interp,var,"max",(char*)ptr->getMax(),0);
	Tcl_SetVar2(interp,var,"low",(char*)ptr->getLow(),0);
	Tcl_SetVar2(interp,var,"high",(char*)ptr->getHigh(),0);
	Tcl_SetVar2(interp,var,"value",(char*)ptr->getValue(img),0);

	coordToTclArray(v, ref, IMAGE, FIXED, var, "Image");
	coordToTclArray(v, ref, PHYSICAL, FIXED, var, "Physical");
	coordToTclArray(v, ref, AMPLIFIER, FIXED, var, "Amplifier");
	coordToTclArray(v, ref, DETECTOR, FIXED, var, "Detector");

	if (ptr->hasWCS()) {
	  char buf[64];
	  ptr->pix2wcs(v*(ptr->*matrixToImage)(), wcs, format, buf, 64);

	  int argc;
	  char** argv;
	  Tcl_SplitList(interp, buf, &argc, &argv);

	  if (argv) {
	    if (argc > 0 && argv[0])
	      Tcl_SetVar2(interp,var,"xWCS",argv[0],0);
	    else
	      Tcl_SetVar2(interp,var,"xWCS","",0);

	    if (argc > 1 && argv[1])
	      Tcl_SetVar2(interp,var,"yWCS",argv[1],0);
	    else
	      Tcl_SetVar2(interp,var,"yWCS","",0);

	    if (argc > 2 && argv[2])
	      Tcl_SetVar2(interp,var,"sysWCS",argv[2],0);
	    else
	      Tcl_SetVar2(interp,var,"sysWCS","",0);

	    if (argc > 3 && argv[3])
	      Tcl_SetVar2(interp,var,"xUnitWCS",argv[3],0);
	    else
	      Tcl_SetVar2(interp,var,"xUnitWCS","",0);

	    if (argc > 4 && argv[4])
	      Tcl_SetVar2(interp,var,"yUnitWCS",argv[4],0);
	    else
	      Tcl_SetVar2(interp,var,"yUnitWCS","",0);

	    Tcl_Free((char*)argv);
	  }
	  else
	    goto noWCS;
	}
	else 
	  goto noWCS;

	return;
      }
      ptr = ptr->next();
    }
  }

  // else, return blanks

  Tcl_SetVar2(interp,var,"objectName","",0);
  Tcl_SetVar2(interp,var,"fileName","",0);
  Tcl_SetVar2(interp,var,"min","",0);
  Tcl_SetVar2(interp,var,"max","",0);
  Tcl_SetVar2(interp,var,"low","",0);
  Tcl_SetVar2(interp,var,"high","",0);
  Tcl_SetVar2(interp,var,"value","",0);
  Tcl_SetVar2(interp,var,"xImage","",0);
  Tcl_SetVar2(interp,var,"yImage","",0);
  Tcl_SetVar2(interp,var,"xPhysical","",0);
  Tcl_SetVar2(interp,var,"yPhysical","",0);
  Tcl_SetVar2(interp,var,"xAmplifier","",0);
  Tcl_SetVar2(interp,var,"yAmplifier","",0);
  Tcl_SetVar2(interp,var,"xDetector","",0);
  Tcl_SetVar2(interp,var,"yDetector","",0);

noWCS:
  Tcl_SetVar2(interp,var,"xWCS","",0);
  Tcl_SetVar2(interp,var,"yWCS","",0);
  Tcl_SetVar2(interp,var,"sysWCS","",0);
  Tcl_SetVar2(interp,var,"xUnitWCS","",0);
  Tcl_SetVar2(interp,var,"yUnitWCS","",0);
}

void Frame::getNameCmd()
{
  Tcl_AppendResult(interp, name, NULL);
}

void Frame::getValueCmd(const Vector& v, CoordSystem ref)
{
  switch (ref) {
  case CANVAS:
    getValue(v, &FitsImage::getCanvasToData);
    break;
  case PANNER:
    getValue(v, &FitsImage::getPannerToData);
    break;
  }
}

void Frame::getValue(const Vector& v, Matrix& (FitsImage::*getMatrix)())
{
  if (fits) {
    FitsImage* ptr = fits;
    while (ptr) {
      Vector img = v * (ptr->*getMatrix)();
      if (img[0]>=0 && img[0]<ptr->width() && 
	  img[1]>=0 && img[1]< ptr->height()) {
	Tcl_AppendResult(interp, ptr->getValue(img), NULL);
	return;
      }
      ptr = ptr->next();
    }
  }
}

void Frame::highliteCmd(int which)
{
  useHighlite = which ? 1 : 0;
  update(PIXMAP);
}

void Frame::setNameCmd(char* n)
{
  if (name)
    delete name;

  name = dupstr(n);
}

void Frame::warpCmd(const Vector& v)
{
  Vector a=v; // can't use a const here
  XWarpPointer(display, None, None, 0, 0, 0, 0, (int)a[0], (int)a[1]);
}

// Clip Commands

void Frame::clipMinMaxModeCmd(MinMaxMode m)
{
  mmMode = m;
  updateClip();
}

void Frame::clipMinMaxParamCmd(int i)
{
  mmIncr = i;
  updateClip();
}

void Frame::clipModeCmd(ClipMode m)
{
  clipMode = m;
  updateClip();
}

void Frame::clipScopeCmd(ClipScope s)
{
  clipScope = s;
  updateClip();
}

void Frame::clipUserCmd(double l, double h)
{
  uLow = l;
  uHigh = h;
  updateClip();
}

void Frame::clipUserLowCmd(double l)
{
  uLow = l;
  updateClip();
}

void Frame::clipUserHighCmd(double h)
{
  uHigh = h;
  updateClip();
}

void Frame::clipZScaleParamCmd(float c, int s, int l)
{
  zContrast = c;
  zSampleSize = s;
  zSampleLine = l;
  updateClip();
}

void Frame::getClipCmd(const Vector& v, CoordSystem ref, Precision p)
{
  switch (ref) {
  case CANVAS:
    getClip(v, &FitsImage::getCanvasToData, p);
    break;
  case PANNER:
    getClip(v, &FitsImage::getPannerToData, p);
    break;
  }
}

void Frame::getClipCmd(Precision p)  // Get Frame Low High
{
  printVectorTcl(Vector(low, high), p);
}

void Frame::getClip(const Vector& v, Matrix& (FitsImage::*getMatrix)(), 
		    Precision p)
{
  if (fits) {
    FitsImage* ptr = fits;
    while (ptr) {
      Vector img = v * (ptr->*getMatrix)();
      if (img[0]>=0 && img[0]<ptr->width() && 
	  img[1]>=0 && img[1]<ptr->height()) {
	printVectorTcl(Vector(ptr->getLowDouble(),ptr->getHighDouble()), p);
	return;
      }
      ptr = ptr->next();
    }
  }
}

void Frame::getClipMinMaxModeCmd()
{
  switch (mmMode) {
  case SCAN:
    Tcl_AppendResult(interp, "scan", NULL);
    return;
  case SAMPLE:
    Tcl_AppendResult(interp, "sample", NULL);
    return;
  case DATAMIN:
    Tcl_AppendResult(interp, "datamin", NULL);
    return;
  case IRAFMIN:
    Tcl_AppendResult(interp, "irafmin", NULL);
    return;
  }
}

void Frame::getClipModeCmd()
{
  switch (clipMode) {
  case MINMAX:
    Tcl_AppendResult(interp, "minmax", NULL);
    break;
  case ZSCALE:
    Tcl_AppendResult(interp, "zscale", NULL);
    break;
  case USERCLIP:
    Tcl_AppendResult(interp, "user", NULL);
    break;
  }
}

void Frame::getClipScopeCmd()
{
  switch (clipScope) {
  case GLOBAL:
    Tcl_AppendResult(interp, "global", NULL);
    break;
  case LOCAL:
    Tcl_AppendResult(interp, "local", NULL);
    break;
  }
}

void Frame::getClipUserCmd(Precision p)
{
  printVectorTcl(Vector(uLow, uHigh), p);
}

void Frame::getClipZScaleParamCmd()
{
  char buf[64];
  ostrstream str(buf,64);
  str << zContrast << ' ' << zSampleSize << ' ' << zSampleLine << ends;

  Tcl_AppendResult(interp, buf, NULL);
}

void Frame::getHistogramCmd(char* x, char* y)
{
  histogram(x,y);
  printDoubleTcl((max-min)/histBins, DEFAULT);
}

void Frame::getMinMaxCmd(const Vector& v, CoordSystem ref)
{
  switch (ref) {
  case CANVAS:
    getMinMax(v, &FitsImage::getCanvasToData);
    break;
  case PANNER:
    getMinMax(v, &FitsImage::getPannerToData);
    break;
  }
}

void Frame::getMinMax(const Vector& v, Matrix& (FitsImage::*getMatrix)())
{
  if (fits) {
    FitsImage* ptr = fits;
    while (ptr) {
      Vector img = v * (ptr->*getMatrix)();
      if (img[0]>= 0 && img[0]<ptr->width() 
	  && img[1]>= 0 && img[1]<ptr->height()) {
	Tcl_AppendElement(interp, (char*)ptr->getMin());
	Tcl_AppendElement(interp, (char*)ptr->getMax());
	return;
      }
      ptr = ptr->next();
    }
  }
}

void Frame::hasDATAMINCmd()
{
  if (currentFits && currentFits->hasDATAMIN())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::hasIRAFMINCmd()
{
  if (currentFits && currentFits->hasIRAFMIN())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

// Crosshair Commands

void Frame::crosshairCmd(int which)
{
  useCrosshair = which ? 1 : 0;
  update(PIXMAP);
}

void Frame::crosshairCmd(const Vector& v, CoordSystem sys)
{
  useCrosshair = 1;
  crosshair = map(v, sys, REF);
  update(PIXMAP);
}

void Frame::crosshairWarpCmd(const Vector& v)
{
  useCrosshair = 1;

  Vector r = crosshair * refToCanvas;
  r += v;
  crosshair = r * canvasToRef;

  update(PIXMAP);
  if (useMagnifierGraphics)
    updateMagnifier();
}

void Frame::crosshairBeginCmd(const Vector& v, CoordSystem sys)
{
  // This routine is optimalized for rendering an interactive cursor. First,
  // a cursorless pixmap is generated. Then as the cursor is moved about, the 
  // cursor is written directly to the display, and erase by XCopyArea from the
  // widget pixmap.

  // turn cursor off and generate pixmap

  useCrosshair = 0;
  updateNow(PIXMAP);

  // turn cursor on, and draw directly to display

  useCrosshair = 1;
  crosshair = map(v, sys, REF);
  drawCrosshair();
}

void Frame::crosshairMotionCmd(const Vector& v, CoordSystem sys)
{
  eraseCrosshair();
  crosshair = map(v, sys, REF);
  drawCrosshair();
}

void Frame::getCrosshairCmd(CoordSystem sys, Precision p)
{
  Vector v = map(crosshair, REF, sys);
  printVectorTcl(v, p);
}

void Frame::getCrosshairStatusCmd()
{
  if (useCrosshair)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::getHorzGraphCmd(char* x, char* y)
{
  graphCut(x, y, Frame::X, crosshair * refToWidget);
}

void Frame::getHorzGraphCmd(char* x, char* y, const Vector& v)
{
  graphCut(x, y, Frame::X, v * canvasToWidget);
}

void Frame::getVertGraphCmd(char* x, char* y)
{
  graphCut(x, y, Frame::Y, crosshair * refToWidget);
}

void Frame::getVertGraphCmd(char* x, char* y, const Vector& v)
{
  graphCut(x, y, Frame::Y, v * canvasToWidget);
}

// Fits Commands

void Frame::getFitsHeaderCmd(int which)
{
  if (which > 0) {
    FitsImage* ptr = fits;
    for (int i=0; i<(which-1); i++) {
      if (ptr)
	ptr = ptr->next();
    }

    if (ptr) {
      char* h = ptr->displayHeader();
      Tcl_AppendResult(interp, h, NULL);
      delete h;
    }
    else {
      char buf[4];
      ostrstream str(buf,4);
      str << which << ends;

      Tcl_AppendResult(interp, " file ", buf, " not loaded.", NULL);
      result = TCL_ERROR;
    }
  }
  else {
    Tcl_AppendResult(interp, " file index must be greater than zero.", NULL);
    result = TCL_ERROR;
  }
}

void Frame::getFitsFileNameCmd(int which, FileNameType type)
{
  if (which > 0) {
    FitsImage* ptr = fits;
    for (int i=0; i<(which-1); i++) {
      if (ptr)
	ptr = ptr->next();
    }

    if (ptr) {
      if (type == Frame::BASIC)
	Tcl_AppendResult(interp, ptr->getFileName(), NULL);
      else
	Tcl_AppendResult(interp, ptr->getFullFileName(), NULL);
    }
    else {
      char buf[4];
      ostrstream str(buf,4);
      str << which << ends;

      Tcl_AppendResult(interp, " file ", buf, " not loaded.", NULL);
      result = TCL_ERROR;
    }
  }
  else {
    Tcl_AppendResult(interp, " file index must be greater than zero.", NULL);
    result = TCL_ERROR;
  }
}

void Frame::getFitsFileNameCmd(const Vector& v, CoordSystem ref, 
			       FileNameType type)
{
  switch (ref) {
  case CANVAS:
    getFitsFileName(v, &FitsImage::getCanvasToData, type);
    break;
  case PANNER:
    getFitsFileName(v, &FitsImage::getPannerToData, type);
    break;
  }
}

void Frame::getFitsFileName(const Vector& v, Matrix& (FitsImage::*getMatrix)(),
			    FileNameType type)
{
  if (fits) {
    FitsImage* ptr = fits;
    while (ptr) {
      Vector img = v * (ptr->*getMatrix)();
      if (img[0]>=0 && img[0]<ptr->width() && 
	  img[1]>=0 && img[1]<ptr->height()) {
	if (type == Frame::BASIC)
	  Tcl_AppendResult(interp, ptr->getFileName(), NULL);
	else
	  Tcl_AppendResult(interp, ptr->getFullFileName(), NULL);
	return;
      }
      ptr = ptr->next();
    }
  }
}

void Frame::getFitsFileNamesCmd(FileNameType type)
{
  FitsImage* ptr = fits;
  while (ptr) {
    if (type == Frame::BASIC)
      Tcl_AppendElement(interp, (char*)ptr->getFileName());
    else
      Tcl_AppendElement(interp, (char*)ptr->getFullFileName());

    ptr = ptr->next();
  }
}

void Frame::getFitsObjectNameCmd(int which)
{
  if (which > 0) {
    FitsImage* ptr = fits;
    for (int i=0; i<(which-1); i++) {
      if (ptr)
	ptr = ptr->next();
    }

    if (ptr)
      Tcl_AppendResult(interp, ptr->getObjectName(), NULL);
    else {
      char buf[4];
      ostrstream str(buf,4);
      str << which << ends;

      Tcl_AppendResult(interp, " file ", buf, " not loaded.", NULL);
      result = TCL_ERROR;
    }
  }
  else {
    Tcl_AppendResult(interp, " file index must be greater than zero.", NULL);
    result = TCL_ERROR;
  }
}

void Frame::getFitsObjectNameCmd(const Vector& v, CoordSystem ref)
{
  switch (ref) {
  case CANVAS:
    getFitsObjectName(v, &FitsImage::getCanvasToData);
    break;
  case PANNER:
    getFitsObjectName(v, &FitsImage::getPannerToData);
    break;
  }
}

void Frame::getFitsObjectName(const Vector& v, 
			      Matrix& (FitsImage::*getMatrix)())
{
  if (fits) {
    FitsImage* ptr = fits;
    while (ptr) {
      Vector img = v * (ptr->*getMatrix)();
      if (img[0]>=0 && 
	  img[0]<ptr->width() && 
	  img[1]>=0 && 
	  img[1]<ptr->height()) {
	const char* n = ptr->getObjectName();
	if (n)
	  Tcl_AppendResult(interp, n, NULL);
	  
	return;
      }
      ptr = ptr->next();
    }
  }
}

void Frame::getFitsMasterCmd()
{
  if (fits) {
    char buf[64];
    ostrstream str(buf,64);
    str << fits->getFullFileName() << ' ' << fits->getFitsFile() << ends;

    Tcl_AppendResult(interp, buf, NULL);
  }
}

void Frame::getFitsSizeCmd(CoordFormat format)
{
  if (currentFits)
    printVectorTcl(mapLenFromRef(currentFits->getSize(), format), FIXED);
  else
    printVectorTcl(Vector(), DEFAULT);
}

void Frame::loadFitsAllocCmd(const char* fn)
{
  FitsImage* img = new FitsImageFitsAlloc(this, fn, 0, 0);
  load(img);
}

void Frame::loadFitsChannelCmd(const char* ch, const char* fn)
{
  FitsImage* img = new FitsImageFitsChannel(this, interp, ch, fn, 0, 0);
  load(img);
}

void Frame::loadFitsMMapCmd(const char* fn)
{
  FitsImage* img = new FitsImageFitsMMap(this, fn, 0, 0);
  load(img);
}

void Frame::loadFitsShareCmd(ShmType type, int id, int size, const char* fn)
{
  FitsImage* img = new FitsImageFitsShare(this, type, id, size, fn, 0, 0);
  load(img);
}

void Frame::loadArrAllocCmd(const char* fn)
{
  FitsImage* img = new FitsImageArrAlloc(this, fn, 0, 0);
  load(img);
}

void Frame::loadArrChannelCmd(const char* ch, const char* fn)
{
  FitsImage* img = new FitsImageArrChannel(this, interp, ch, fn, 0, 0);
  load(img);
}

void Frame::loadArrMMapCmd(const char* fn)
{
  FitsImage* img = new FitsImageArrMMap(this, fn, 0, 0);
  load(img);
}

void Frame::loadArrShareCmd(ShmType type, int id, int size, const char* fn)
{
  FitsImage* img = new FitsImageArrShare(this, type, id, size, fn, 0, 0);
  load(img);
}

void Frame::loadFitsSlaveCmd(const char* fn, FitsFile* f)
{
  FitsImage* img = new FitsImageSlave(this,fn,0,0,f);
  load(img);
}

void Frame::load(FitsImage* img)
{
  unloadFits();

  if (!img->isValid()) {
    delete img;

    reset();
    updateClip();
    update(MATRIX);
    
    Tcl_AppendResult(interp, "Unable to load file", NULL);
    result = TCL_ERROR;
    return;
  }

  fits = img;
  fitsCount++;
  currentFits = fits;

  // reset some values

  centerImage();
  crosshair = cursor;
  wcsAlign();

  // update clip values

  clipScope = LOCAL;
  
  if (!fits->hasDATAMIN() && mmMode == DATAMIN)
    mmMode = SAMPLE;
  if (!fits->hasIRAFMIN() && mmMode == IRAFMIN)
    mmMode = SAMPLE;

  updateClip();

  // and schedule an update event

  update(MATRIX);
}

void Frame::loadMosaicExtMMapCmd(const char* fn)
{
  FitsImage* img = new FitsImageMosaicMMap(this, fn);
  loadMosaicExt(MOSAICMAP,fn,img);
}

void Frame::loadMosaicExtShareCmd(ShmType type, int id, int size,
				  const char* fn)
{
  FitsImage* img = new FitsImageMosaicShare(this, type, id, size, fn);
  loadMosaicExt(MOSAICMAP,fn,img);
}

void Frame::loadMosaicExtAllocCmd(const char* fn)
{
  FitsImage* img = new FitsImageMosaicAlloc(this, fn);
  loadMosaicExt(MOSAICALLOC,fn,img);
}

void Frame::loadMosaicExtChannelCmd(const char* ch, const char* fn)
{
  FitsImage* img = new FitsImageMosaicChannel(this, interp, ch, fn);
  loadMosaicExt(MOSAICCHANNEL,fn,img);
}

void Frame::loadMosaicExt(MosaicExtType which, const char* fn, FitsImage* img)
{
  if (!img->isValid() || !img->isMosaic()) {
    delete img;
    
    Tcl_AppendResult(interp, "Unable to load mosaic file", NULL);
    result = TCL_ERROR;
    return;
  }

  // Its legal, save it

  unloadFits();

  fits = img;
  currentFits = img; // the first is our 'current'
  fitsCount=1;

  // get the rest

  FitsImage* ptr = fits;
  while (1) {
    FitsImage* next = loadMosaicExtNext(which, fn, ptr->getFitsFile());
    if (next->isValid()) {
      if (next->isImage() && next->isMosaic()) {
	ptr->setNext(next);
	ptr = next;
	fitsCount++;

	// reset clip mode if needed

	if (!next->hasDATAMIN() && mmMode == DATAMIN)
	  mmMode = SAMPLE;
	if (!next->hasIRAFMIN() && mmMode == IRAFMIN)
	  mmMode = SAMPLE;
      }
      else
	// is not a mosaic image, skip to next ext
	delete next;
    }
    else {
      // we must be at the end, or an error has occured
      delete next;
      break;
    }
  }

  // reset some values
  
  centerImage();
  crosshair = cursor;
  wcsAlign();

  // update clip values

  updateClip();

  // and schedule an update event

  update(MATRIX);
}

FitsImage* Frame::loadMosaicExtNext(MosaicExtType which, const char* fn, 
				    FitsFile* ff)
{
  switch (which) {
  case MOSAICMAP:
     return new FitsImageMosaicNextMap(this, fn, ff);
  case MOSAICALLOC:
    return new FitsImageMosaicNextAlloc(this, fn, ff);
  case MOSAICCHANNEL:
    return new FitsImageMosaicNextChannel(this, fn, ff);
  }
}

void Frame::loadMosaicAllocCmd(const char* fn)
{
  FitsImage* img = new FitsImageFitsAlloc(this, fn, 1, !fitsCount);
  loadMosaic(img);
}

void Frame::loadMosaicChannelCmd(const char* ch, const char* fn)
{
  FitsImage* img = new FitsImageFitsChannel(this, interp, ch, fn, 1, 
					    !fitsCount);
  loadMosaic(img);
}

void Frame::loadMosaicMMapCmd(const char* fn)
{
  FitsImage* img = new FitsImageFitsMMap(this, fn, 1, !fitsCount);
  loadMosaic(img);
}

void Frame::loadMosaicShareCmd(ShmType type, int id, int size, const char* fn)
{
  FitsImage* img = new FitsImageFitsShare(this, type, id, size, fn, 1, 
					  !fitsCount);
  loadMosaic(img);
}

void Frame::loadMosaic(FitsImage* img)
{
  if (!img->isValid() || !img->isMosaic()) {
    delete img;
    
    Tcl_AppendResult(interp, "Unable to load mosaic file", NULL);
    result = TCL_ERROR;
    return;
  }

  // Its legal, save it

  fitsCount++;

  if (fits) {
    FitsImage* ptr = fits;
    while (ptr && ptr->next())
      ptr = ptr->next();
    ptr->setNext(img);
  }
  else {
    fits = img;
    currentFits = img; // the first is our 'current'

    // reset some values

    centerImage();
    crosshair = cursor;
    wcsAlign();
  }

  // update clip values

  if (!fits->hasDATAMIN() && mmMode == DATAMIN)
    mmMode = SAMPLE;
  if (!fits->hasIRAFMIN() && mmMode == IRAFMIN)
    mmMode = SAMPLE;

  updateClip();

  // and schedule an update event

  update(MATRIX);
}

void Frame::saveFitsFileCmd(const char* fn)
{
  if (fits) {
    if (!fits->saveFitsFile(fn)) {
      Tcl_AppendResult(interp, "Unable to save Fits", NULL);
      result = TCL_ERROR;
    }
  }
}

void Frame::saveFitsChannelCmd(const char* ch)
{
  if (fits) {
    if (!fits->saveFitsChannel(interp, ch)) {
      Tcl_AppendResult(interp, "Unable to save fits to channel ", ch, NULL);
      result = TCL_ERROR;
    }
  }
}

void Frame::hasFitsCmd()
{
  if (fitsCount)
      Tcl_AppendResult(interp, "1", NULL);
  else
      Tcl_AppendResult(interp, "0", NULL);
}

void Frame::updateFitsCmd(int which, BBox bb)
{
  // Note: bb is in IMAGE coords

  FitsImage* ptr = fits;
  if (which > 0) {
    for (int i=0; i<(which-1); i++) {
      if (ptr)
	ptr = ptr->next();
    }
  }

  if (ptr) {
    Vector ll = bb.ll * ptr->getImageToCanvas();
    Vector lr = bb.lr() * ptr->getImageToCanvas();
    Vector ur = bb.ur * ptr->getImageToCanvas();
    Vector ul = bb.ul() * ptr->getImageToCanvas();
    BBox rr(ll);
    rr.bound(lr);
    rr.bound(ur);
    rr.bound(ul);
    updateNow(BASE, rr);
  }
}

void Frame::unloadFitsCmd()
{
  unloadFits();

  // reset some values

  centerImage();
  wcsAlign();
  crosshair = cursor;

  update(MATRIX);
}

void Frame::scanOnLoadCmd(int which)
{
  scanOnLoad = which;
}

// Coordinate Commands

void Frame::getWCSCoordCmd(const Vector& v, CoordSystem in, 
			   CoordSystem wcs, CoordFormat format)
{
  char buf[64];
  strcpy(buf, "");
  mapToWCSStr(v, in, wcs, format, buf, 64);
  Tcl_AppendResult(interp, buf, NULL);
}

void Frame::getWCSFlipCmd()
{
  if (currentFits && currentFits->getWCSFlip())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::getWCSSystemCmd()
{
  if (currentFits)
    Tcl_AppendResult(interp, currentFits->getWCSSystem(), NULL);
}

void Frame::hasMosaicCmd()
{
  if (currentFits && currentFits->isMosaic())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::hasPhysicalCmd()
{
  if (currentFits && currentFits->hasLTMV())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::hasDetectorCmd()
{
  if (currentFits && currentFits->hasDTMV())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::hasAmplifierCmd()
{
  if (currentFits && currentFits->hasATMV())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::hasWCSCmd()
{
  if (currentFits && currentFits->hasWCS())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::hasWCSEquCmd()
{
  if (currentFits && currentFits->hasWCS() && currentFits->hasWCSEqu())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::hasWCSLinearCmd()
{
  if (currentFits && currentFits->hasWCS() && currentFits->hasWCSLinear())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

// Pan Zoom Commands

void Frame::centerCmd()
{
  centerImage();
  update(MATRIX);
}

void Frame::getPanCmd(CoordSystem sys, Precision p)
{
  Vector v = map(cursor, USER, sys);
  printVectorTcl(v, p);
}

void Frame::getOrientCmd()
{
  switch (orientation) {
  case NORMAL:
    Tcl_AppendResult(interp, "none", NULL);
    return;
  case X:
    Tcl_AppendResult(interp, "x", NULL);
    return;
  case Y:
    Tcl_AppendResult(interp, "y", NULL);
    return;
  case XY:
    Tcl_AppendResult(interp, "xy", NULL);
    return;
  }
}

void Frame::getRotateCmd(Precision p)
{
  printDoubleTcl(radToDeg(rotation), p);
}

void Frame::getWCSAlignCmd()
{
  if (wcsAlignment)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::getWCSZoomCmd(Precision p)
{
  if (currentFits && currentFits->hasWCS() && currentFits->hasWCSEqu()) {
    Vector v = currentFits->getWCScdelt();
    printDoubleTcl(v[0], p);
  }
  else
    Tcl_AppendResult(interp, "1", NULL);
}

void Frame::getZoomCmd(Precision p)
{
  printDoubleTcl(zoom, p);
}

void Frame::orientCmd(Frame::Orientation which)
{
  orientation = which;

  switch (orientation) {
  case NORMAL:
    orientationMatrix.identity();
    break;
  case X:
    orientationMatrix = FlipX();
    break;
  case Y:
    orientationMatrix = FlipY();
    break;
  case XY:
    orientationMatrix = FlipXY();
    break;
  }

  update(MATRIX);
}

void Frame::panCmd(const Vector& v1, const Vector& v2)
{
  Vector start = v1 * canvasToUser;
  Vector stop = v2 * canvasToUser;
  cursor -= stop - start;
  update(MATRIX);
}

void Frame::panCmd(const Vector& v, CoordSystem sys)
{
  int valid;
  Vector u = map(cursor, USER, sys, &valid);
  if (valid) {
    u += v;
    cursor = map(u, sys, USER);
    if (currentFits)
      currentFits->setHistCursor(cursor);
    update(MATRIX);
  }
  else {
    Tcl_AppendResult(interp, "Invalid Coordinate", NULL);
    result = TCL_ERROR;
  }
}

void Frame::panBBoxCmd(BBox bb)
{
  if (currentFits) {       
    // we want to round to nearest pixel center
    Vector r = bb.center() * currentFits->getPannerToData();
    cursor = (r.floor() + Vector(.5,.5)) * fits->getDataToUser();
    if (currentFits)
      currentFits->setHistCursor(cursor);
    update(MATRIX);
  }
}

void Frame::panMotionBeginCmd(const Vector& v)
{
  // save the current cursor

  panCursor = v;

  // copy tmp pixmap

  panPM = Tk_GetPixmap(display, Tk_WindowId(tkwin),
		       options->width, options->height, depth);
  if (!panPM) {
    cerr << "Frame Internal Error: Unable to Create Pan Motion Pixmap" << endl;
    exit(1);
  }
  XCopyArea(display, pixmap, panPM, gc, 0, 0, options->width, 
	    options->height, 0,0);
}

void Frame::panMotionCmd(const Vector& v)
{
  if (visible) {

    // Clear

    Vector diff = ((v * canvasToWidget) - (panCursor * canvasToWidget));

    BBox hh,vv;
    if (diff[0]>0 && diff[1]>0) {
      hh = BBox(Vector(0,0), Vector(options->width, diff[1]));
      vv = BBox(Vector(0,0), Vector(diff[0], options->height));
    } else if (diff[0]>0 && diff[1]<0) {
      hh = BBox(Vector(options->width,options->height), 
		Vector(0,options->height+diff[1]));
      vv = BBox(Vector(0,0), Vector(diff[0], options->height));
    } else if (diff[0]<0 && diff[1]>0) {
      hh = BBox(Vector(0,0), Vector(options->width, diff[1]));
      vv = BBox(Vector(options->width,options->height), 
		Vector(options->width+diff[0], 0));
    } else if (diff[0]<0 && diff[1]<0) {
      hh = BBox(Vector(options->width,options->height), 
		Vector(0,options->height+diff[1]));
      vv = BBox(Vector(options->width,options->height), 
		Vector(options->width+diff[0], 0));
    }

    hh = hh * widgetToWindow;
    vv = vv * widgetToWindow;
    
    Vector hs = hh.size();
    XClearArea(display, Tk_WindowId(tkwin),
	       (int)hh.ll[0], (int)hh.ll[1], (int)hs[0], (int)hs[1], False);

    Vector vs = vv.size();
    XClearArea(display, Tk_WindowId(tkwin),
	       (int)vv.ll[0], (int)vv.ll[1], (int)vs[0], (int)vs[1], False);

    // display tmp pixmap at new location

    Vector d = ((v * canvasToWidget) - (panCursor * canvasToWidget)) * 
      widgetToWindow;
    XCopyArea(display, panPM, Tk_WindowId(tkwin), panGCXOR, 
	      0, 0, options->width, options->height, (int)d[0], (int)d[1]);
  }
}

void Frame::panMotionEndCmd(const Vector& v)
{
  // delete tmp pixmap

  if (panPM)
    Tk_FreePixmap(display, panPM);
  panPM = 0;

  Vector start = panCursor * canvasToUser;
  Vector stop = v * canvasToUser;
  cursor -= stop - start;
  update(MATRIX);
}

void Frame::panToCmd(const Vector& v, CoordSystem sys)
{
  int valid;
  Vector vv = map(v, sys, USER, &valid);
  if (valid) {
    cursor = vv;
    if (currentFits)
      currentFits->setHistCursor(cursor);
    update(MATRIX);
  }
  else {
    Tcl_AppendResult(interp, "Invalid Coordinate", NULL);
    result = TCL_ERROR;
  }
}

void Frame::rotateCmd(double r)
{
  rotation += r;
  update(MATRIX);
}

void Frame::rotateToCmd(double r)
{
  rotation = r;
  update(MATRIX);
}

void Frame::rotateMotionBeginCmd()
{
  // save the current rotation

  rotateRotation = rotation;

  // Create src XImage

  if (!(rotateSrcXM = XGetImage(display, pixmap, 0, 0,
				options->width, options->height, 
				AllPlanes, ZPixmap))) {
    cerr << "Frame Internal Error: Unable to Create Rotate XImage" << endl;
    exit(1);
  }

  // Create dest XImage

  if (!(rotateDestXM = XGetImage(display, pixmap, 0, 0,
				options->width, options->height, 
				 AllPlanes, ZPixmap))) {
    cerr << "Frame Internal Error: Unable to Create Rotate XImage" << endl;
    exit(1);
  }

  // Create dest Pixmap

  rotatePM = Tk_GetPixmap(display, Tk_WindowId(tkwin), 
			  options->width, options->height, depth);
  if (!rotatePM) {
    cerr << "Frame Internal Error: Unable to Create Rotate Motion Pixmap" 
	 << endl;
    exit(1);
  }
}

void Frame::rotateMotionCmd(double angle)
{
  rotation = rotateRotation + angle;
  rotateMotion();
}

void Frame::rotateMotionEndCmd()
{
  // Clean up

  if (rotateSrcXM)
    XDestroyImage(rotateSrcXM);
 
  if (rotateDestXM)
    XDestroyImage(rotateDestXM);

  if (rotatePM)
    Tk_FreePixmap(display, rotatePM);

  update(MATRIX);
}

void Frame::wcsAlignCmd(int which)
{
  wcsAlignment = which;
  wcsAlign();
  update(MATRIX);
}

void Frame::wcsZoomCmd(double a)
{
  if (currentFits && currentFits->hasWCS() && currentFits->hasWCSEqu()) {
    Vector b = currentFits->getWCScdelt();
    wcsZoom = fabs(b[0]/a);
  }
  else
    wcsZoom = 1;

  update(MATRIX);
}

void Frame::zoomCmd(double z)
{
  zoom *= z;
  update(MATRIX);
}

void Frame::zoomAboutCmd(double z, const Vector& v, CoordSystem sys)
{
  zoom *= z;
  int valid;
  Vector vv = map(v, sys, USER, &valid);
  if (valid) {
    cursor = vv;
    if (currentFits)
      currentFits->setHistCursor(cursor);
    update(MATRIX);
  }
  else {
    Tcl_AppendResult(interp, "Invalid Coordinate", NULL);
    result = TCL_ERROR;
  }
}

void Frame::zoomToCmd(double z)
{
  zoom = z;
  update(MATRIX);
}

void Frame::zoomToAboutCmd(double z, const Vector& v, CoordSystem sys)
{
  zoom = z;
  int valid;
  Vector vv = map(v, sys, USER, &valid);
  if (valid) {
    cursor = vv;
    if (currentFits)
      currentFits->setHistCursor(cursor);
    update(MATRIX);
  }
  else {
    Tcl_AppendResult(interp, "Invalid Coordinate", NULL);
    result = TCL_ERROR;
  }
}

void Frame::zoomToFitCmd()
{
  if (fits) {
    centerImage();
    zoom = calcZoom(options->width,options->height,fits->getSize());
    update(MATRIX);
  }
}

// Panner/Magnifer Commands

void Frame::magnifierCmd(char* n, int w, int h)
{
  strcpy(magnifierName,n);
  magnifierWidth = w;
  magnifierHeight = h;

  if (magnifierPixmap)
    Tk_FreePixmap(display, magnifierPixmap);
  magnifierPixmap = 0;

  if (magnifierXImage)
    XDestroyImage(magnifierXImage);
  magnifierXImage = NULL;

  if (magnifierWidth > 0 && magnifierHeight > 0) {
    magnifierPixmap = Tk_GetPixmap(display, Tk_WindowId(tkwin),
			     magnifierWidth, magnifierHeight, depth);
    if (!magnifierPixmap) {
      cerr << "Frame Internal Error: Unable to Create Magnifier Pixmap" 
	   << endl;
      exit(1);
    }

    if (!magnifierXImage)
      if (!(magnifierXImage = 
	    XGetImage(display, magnifierPixmap, 0, 0, 
		      magnifierWidth, magnifierHeight,
		      AllPlanes, ZPixmap))){
	cerr << "Frame Internal Error: Unable to Create Magnifier XImage" 
	     << endl;
	exit(1);
      }
  }

  if (!magnifierGC)
    magnifierGC = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);

  updateMagnifierMatrix();
}

void Frame::magnifierCmd(int s)
{
  useMagnifier = s;
}

void Frame::magnifierGraphicsCmd(int which)
{
  useMagnifierGraphics = which;
  if (useMagnifier)
    updateMagnifier();
}

void Frame::magnifierCursorCmd(int which)
{
  useMagnifierCursor = which;
  if (useMagnifier)
    updateMagnifier();
}

void Frame::magnifierModeCmd(MagnifierMode m)
{
  magnifierMode = m;
  if (useMagnifier)
    updateMagnifier();
}

void Frame::magnifierZoomCmd(double z)
{
  magnifierZoom = z;
  updateMagnifierMatrix();
  if (useMagnifier)
    updateMagnifier();
}

void Frame::pannerCmd(char* n, int w, int h)
{
  strcpy(pannerName,n);
  pannerWidth = w;
  pannerHeight = h;

  if (pannerPixmap)
    Tk_FreePixmap(display, pannerPixmap);
  pannerPixmap = 0;

  if (pannerXImage)
    XDestroyImage(pannerXImage);
  pannerXImage = NULL;

  if (pannerWidth > 0 && pannerHeight > 0) {
    if (!(pannerPixmap = Tk_GetPixmap(display, Tk_WindowId(tkwin), 
				      pannerWidth, pannerHeight, depth))) {
      cerr << "Frame Internal Error: Unable to Create Panner Pixmap" << endl;
      exit(1);
    }

    if (!(pannerXImage = 
  	  XGetImage(display, pannerPixmap, 0, 0, pannerWidth, pannerHeight,
		    AllPlanes, ZPixmap))){
      cerr << "Frame Internal Error: Unable to Create Panner XImage" << endl;
      exit(1);
    }
  }

  update(MATRIX); // update panner matrices
}

void Frame::pannerCmd(int s)
{
  usePanner = s;
  updatePanner();
}

void Frame::updatePannerCmd()
{
  pannerPush();
}

// Bin Commands

void Frame::binCmd(double b, const Vector& v, CoordSystem sys, 
		   const char* x, const char* y, const char* filter)
{
  binFactor = b;
  if (binFactor <= 0)
    binFactor = 1;
  if (currentFits) {
    currentFits->setHistX(x);
    currentFits->setHistY(y);
    currentFits->setHistFilter(filter);
    Matrix m = currentFits->updateHist(map(v, sys, USER));
    updateMarkerCoords(m);
    if (contour)
      contour->updateCoords(m);
  }

  updateClip();
  update(MATRIX); // because we have changed zoom
}

void Frame::binCmd(double b, const char* x, const char* y, const char* filter)
{
  binFactor = b;
  if (binFactor <= 0)
    binFactor = 1;
  if (currentFits) {
    currentFits->setHistX(x);
    currentFits->setHistY(y);
    currentFits->setHistFilter(filter);
    Matrix m = currentFits->updateHistCenter();
    updateMarkerCoords(m);
    if (contour)
      contour->updateCoords(m);
  }

  updateClip();
  update(MATRIX); // because we have changed zoom
}

void Frame::binAboutCmd(const Vector& v, CoordSystem sys)
{
  if (currentFits) {
    Matrix m = currentFits->updateHist(map(v, sys, USER));
    updateMarkerCoords(m);
    if (contour)
      contour->updateCoords(m);
  }

  updateClip();
  update(MATRIX); // because we have changed zoom
}

void Frame::binColsCmd(const char* x, const char* y)
{
  if (currentFits) {
    currentFits->setHistX(x);
    currentFits->setHistY(y);
    Matrix m = currentFits->updateHistCursor();
    updateMarkerCoords(m);
    if (contour)
      contour->updateCoords(m);
  }

  updateClip();
  update(MATRIX);
}

void Frame::binFactorCmd(double b)
{
  binFactor *= b;
  if (binFactor <= 0)
    binFactor = 1;
  if (currentFits) {
    Matrix m = currentFits->updateHistCursor();
    updateMarkerCoords(m);
    if (contour)
      contour->updateCoords(m);
  }

  updateClip();
  update(MATRIX); // because we have changed zoom
}

void Frame::binFactorAboutCmd(double b, const Vector& v, CoordSystem sys)
{
  binFactor *= b;
  if (binFactor <= 0)
    binFactor = 1;
  if (currentFits) {
    Matrix m = currentFits->updateHist(map(v, sys, USER));
    updateMarkerCoords(m);
    if (contour)
      contour->updateCoords(m);
  }

  updateClip();
  update(MATRIX); // because we have changed zoom
}

void Frame::binFactorToCmd(double b)
{
  binFactor = b;
  if (binFactor <= 0)
    binFactor = 1;
  if (currentFits) {
    Matrix m = currentFits->updateHistCursor();
    updateMarkerCoords(m);
    if (contour)
      contour->updateCoords(m);
  }

  updateClip();
  update(MATRIX); // because we have changed zoom
}

void Frame::binFactorToAboutCmd(double b, const Vector& v, CoordSystem sys)
{
  binFactor = b;
  if (binFactor <= 0)
    binFactor = 1;
  if (currentFits) {
    Matrix m = currentFits->updateHist(map(v, sys, USER));
    updateMarkerCoords(m);
    if (contour)
      contour->updateCoords(m);
  }

  updateClip();
  update(MATRIX); // because we have changed zoom
}

void Frame::binToFitCmd()
{
  if (currentFits) {
    double bf = 1/calcZoom(options->width,options->height,
			   currentFits->getHistDim());

    // round up to next power of 2

    if (bf < 1)
      binFactor = 1;
    else {
      int p=1;
      while (p<bf)
	p*=2;
      binFactor = p;
    }
  
    Matrix m = currentFits->updateHistCursor();
    updateMarkerCoords(m);
    if (contour)
      contour->updateCoords(m);
  }

  updateClip();
  update(MATRIX); // because we have changed zoom
}

void Frame::binFunctionCmd(FitsHist::Function func)
{
  binFunction = func;
  if (currentFits) {
    Matrix m = currentFits->updateHistCursor();
    updateMarkerCoords(m);
    if (contour)
      contour->updateCoords(m);
  }

  updateClip();
  update(MATRIX);
}

void Frame::binBufferSizeCmd(int size)
{
  binBufferSize = size;
  if (currentFits) {
    Matrix m = currentFits->updateHistCursor();
    updateMarkerCoords(m);
    if (contour)
      contour->updateCoords(m);
  }

  updateClip();
  update(MATRIX);
}

void Frame::binFilterCmd(const char* filter)
{
  if (currentFits) {
    currentFits->setHistFilter(filter);
    Matrix m = currentFits->updateHistCursor();
    updateMarkerCoords(m);
    if (contour)
      contour->updateCoords(m);
  }

  updateClip();
  update(MATRIX);
}

void Frame::getBinCursorCmd(Precision p)
{
  if (currentFits)
    printVectorTcl(currentFits->getHistCursor(), p);
  else
    printVectorTcl(Vector(), p);
}

void Frame::getBinFactorCmd(Precision p)
{
  printDoubleTcl(binFactor, p);
}

void Frame::getBinFunctionCmd()
{
  switch (binFunction) {
  case FitsHist::AVERAGE:
    Tcl_AppendResult(interp, "average", NULL);
    return;
  case FitsHist::SUM:
    Tcl_AppendResult(interp, "sum", NULL);
    return;
  }
}

void Frame::getBinFilterCmd()
{
  if (currentFits && currentFits->isHist())
    Tcl_AppendResult(interp, currentFits->getHistFilter(), NULL);
  else
    Tcl_AppendResult(interp, "", NULL);
}

void Frame::getBinColsCmd()
{
  if (currentFits && currentFits->isHist())
    Tcl_AppendResult(interp, currentFits->getHistX(), " ", 
		     currentFits->getHistY(), NULL);
  else
    Tcl_AppendResult(interp, "", NULL);
}

void Frame::getBinListCmd()
{
  if (currentFits && currentFits->isHist()) {
    char* cols = currentFits->getHistList();
    Tcl_AppendResult(interp, cols, NULL);
    delete cols;
  }
  else
    Tcl_AppendResult(interp, "", NULL);
}

void Frame::hasBinCmd()
{
  if (currentFits && currentFits->isHist())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::contourCreateCmd(const char* color, int num, float res, 
			     ContourMethod al)
{
  if (contour)
    delete contour;
  contour = NULL;

  if (fits) {
    if (colorScaleType != IISSCALE) 
      contour = new FVContourPlot(this, fits, color, num, res, al, low, high,
				  colorScaleType);
    else
      contour = new FVContourPlot(this, fits, color, num, res, al, 0, 199,
				  LINEARSCALE);

    FitsImage* ptr = fits->next();
    while (ptr) {
      ((FVContourPlot*)contour)->append(ptr);
      ptr = ptr->next();
    }
  }

  update(BASE);
}

void Frame::contourCreateCmd(const char* color, int num, float res,
			     ContourMethod al, ColorScaleType s)
{
  if (contour)
    delete contour;
  contour = NULL;

  if (fits) {
    contour = new FVContourPlot(this, fits, color, num, res, al, low, high, s);
    FitsImage* ptr = fits->next();
    while (ptr) {
      ((FVContourPlot*)contour)->append(ptr);
      ptr = ptr->next();
    }
  }

  update(BASE);
}

void Frame::contourCreateCmd(const char* color, int num, float res, 
			     ContourMethod al,
			     float l, float h, ColorScaleType s)
{
  if (contour)
    delete contour;
  contour = NULL;

  if (fits) {
    contour = new FVContourPlot(this, fits, color, num, res, al, l, h, s);
    FitsImage* ptr = fits->next();
    while (ptr) {
      ((FVContourPlot*)contour)->append(ptr);
      ptr = ptr->next();
    }
  }

  update(BASE);
}

void Frame::contourCreateCmd(const char* color, float res, ContourMethod al,
			     const char* lev)
{
  if (contour)
    delete contour;
  contour = NULL;

  int cnt = 0;
  float levels[100];
  istrstream str(lev);
  while ((cnt<100) && (str >> levels[cnt]))
    cnt++;

  cnt -= 1; // -1 for diffs in levels

  if (fits) {
    contour = new FVContourPlot(this, fits, color, cnt, res, al, levels);
    FitsImage* ptr = fits->next();
    while (ptr) {
      ((FVContourPlot*)contour)->append(ptr);
      ptr = ptr->next();
    }
  }

  update(BASE);
}

void Frame::contourLoadCmd(const char* color, const char* fn,
			     CoordSystem sys)
{
  if (contour)
    delete contour;
  contour = NULL;

  ifstream str(fn);

  if (str) {
    List<Vertex>* c = new List<Vertex>;
 
    int valid;
    while (!str.eof()) {
      Vertex* vv = new Vertex();

      char buf[64];
      str.getline(buf,64,'\n');
      if (strlen(buf) > 0) {
	Vector v;
	istrstream sstr(buf);
	sstr >> v[0] >> v[1];
	vv->vector = map(v, sys, REF, &valid);
      }
      else
	vv->vector = Vector(DBL_MAX, DBL_MAX);

      c->append(vv);
    }

    contour = new ContourPlot(this, color, c);
  }

  update(BASE);
}

void Frame::contourDeleteCmd()
{
  if (contour)
    delete contour;
  contour = NULL;

  update(BASE);
}

void Frame::contourCopyCmd(CoordSystem sys)
{
  if (contour) {
    List<Vertex>* v = contour->dup();
    int valid;
    if (v->head())
      do {
	Vector& w = v->current()->vector;
	if (w[0] != DBL_MAX)
	  w = map(w, REF, sys, &valid);
      } while (v->next());

    char buf[32];
    ostrstream str(buf,32);
    str << hex << v << ends;
    Tcl_AppendResult(interp, buf, NULL);
  }
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::contourPasteCmd(const char* c, void* u, CoordSystem sys)
{
  List<Vertex>* v = (List<Vertex>*)u;

  if (contour)
    delete contour;

  int valid;
  if (v->head())
      do {
	Vector& w = v->current()->vector;
	if (w[0] != DBL_MAX)
	  w = map(w, sys, REF, &valid);
      } while (v->next());
  
  contour = new ContourPlot(this, c, v);

  update(BASE);
}

void Frame::contourSaveCmd(const char* fn, CoordSystem sys)
{
  if (contour) {
    ofstream str(fn);

    if (str) {
      List<Vertex>* c = contour->getContours();

      int valid = 1;
      if (c->head())
	do {
	  Vector v = (c->current())->vector;
	  if (v[0] != DBL_MAX && v[1] != DBL_MAX)
	    str << setiosflags(ios::scientific) 
		<< setprecision(8) 
		<< map(v, REF, sys, &valid) 
		<< endl;
	  else
	    str << endl;
	}
	while (c->next() && valid);
    }
  }
}

void Frame::getContourLevelCmd(int count, float ll, float hh,
			       ColorScaleType scaleType)
{
  ContourScale* scale;
  
  switch (scaleType) {
  case Frame::LINEARSCALE:
    scale = new LinearContourScale(count-1, ll, hh);
    break;
  case Frame::LOGSCALE:
    scale = new LogContourScale(count-1, ll, hh);
    break;
  case Frame::SQUAREDSCALE:
    scale = new SquaredContourScale(count-1, ll, hh);
    break;
  case Frame::SQRTSCALE:
    scale = new SqrtContourScale(count-1, ll, hh);
    break;
  }

  char buf[2048];
  ostrstream str(buf,2048);
  str << *scale << ends;

  Tcl_AppendResult(interp, buf, NULL);

  delete scale;
}

void Frame::getContourMethodCmd()
{
  if (contour)
    switch ((Frame::ContourMethod)(((FVContourPlot*)contour)->getMethod())) {
    case Frame::AVERAGE:
      Tcl_AppendResult(interp, "average", NULL);
      break;
    case Frame::MAX:
      Tcl_AppendResult(interp, "max", NULL);
      break;
    }
  else
    Tcl_AppendResult(interp, "", NULL);
}

void Frame::hasContourCmd()
{
  if (contour)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

// IIS Commands

void Frame::iisCmd(int width, int height)
{
  unloadFits();

  // force clip mode to user between 1 to 200

  clipMode = USERCLIP;
  mmMode = SAMPLE;
  mmIncr = 1;
  uLow = IISMIN;
  uHigh = IISMAX;
  
  colorScaleType = IISSCALE;
  updateColorScale();

  FitsImage* img = new FitsImageIIS(this, width, height);

  load(img);
}

void Frame::iisEraseCmd()
{
  if (fits)
    ((FitsImageIIS*)fits)->iisErase();
}

void Frame::iisMessageCmd(const char* ptr)
{
  if (fits) {
    char fn[64];
    istrstream str(ptr);
    str >> fn >> fn;
    fits->setFileName(fn);
  }
}

void Frame::iisGetCmd(char* ptr, int x, int y, int dx, int dy)
{
  if (fits)
    memcpy(ptr, ((FitsImageIIS*)fits)->iisGet(x,y), dx*dy);
}

void Frame::iisGetWidthCmd()
{
  if (fits)
    printIntegerTcl(fits->width());
}

void Frame::iisGetHeightCmd()
{
  if (fits)
    printIntegerTcl(fits->height());
}

void Frame::iisSetCmd(const char* data, int x, int y, int dx, int dy)
{
  if (fits)
    ((FitsImageIIS*)fits)->iisSet(x, y, data, dx*dy);
}

void Frame::iisWCSCmd(const Matrix& m, const Vector& z, int zt)
{
  if (fits)
    ((FitsImageIIS*)fits)->iisWCS(m, z, zt);
}

void Frame::iisSetCursorCmd(const Vector& v, CoordSystem sys)
{
  iisLastCursor = map(v,sys,WIDGET);

  Vector r = iisLastCursor * widgetToWindow;
  XWarpPointer(display, None, Tk_WindowId(tkwin), 0,0,0,0,(int)r[0],(int)r[1]);
}

void Frame::iisGetCursorCmd()
{
  printVectorTcl(iisLastCursor,DEFAULT);
}

void Frame::iisCursorModeCmd(int state)
{
  if (state) {
    // save cursor pos

    Window root, child;
    int rootx,rooty,winx,winy;
    unsigned int mask;
    XQueryPointer(display,Tk_WindowId(tkwin),&root,&child,
		  &rootx,&rooty,&winx,&winy,&mask);

    if (!iisInteractive) {
      iisSaveRoot = root;
      iisSaveCursor = Vector(rootx,rooty);
    }
    iisInteractive++;

    // set iisLastCursor if first time thru
    if (iisLastCursor[0] == 0 && iisLastCursor[1] == 0)
      iisLastCursor = Vector(options->width,options->height)/2;

    // and move to last cursor position
    Vector r = iisLastCursor * widgetToWindow;
    XWarpPointer(display, None, Tk_WindowId(tkwin),0,0,0,0,(int)r[0],(int)r[1]);
  }
  else {
    // restore cursor pos

    if (!(--iisInteractive)) {
      XWarpPointer (display, None, iisSaveRoot, 0, 0, 0, 0, 
		    (int)iisSaveCursor[0], (int)iisSaveCursor[1]);
      iisSaveRoot = 0;
      iisSaveCursor = Vector();
    }
  }
}

void Frame::hasIISCmd()
{
  if (isIIS())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

// Grid Commands

void Frame::gridCmd(CoordSystem sys, const char* ops, int ext)
{
  if (grid)
    delete grid;

  grid = new Grid(this, sys, ops, ext);

  update(BASE);
}

void Frame::gridDeleteCmd()
{
  if (grid)
    delete grid;
  grid = NULL;

  update(BASE);
}

void Frame::hasGridCmd()
{
  if (grid)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::getGridSystemCmd()
{
  if (grid) {
    switch((Frame::CoordSystem)grid->system()) {
    case Frame::EQUATORIAL_FK4:
      Tcl_AppendResult(interp, "fk4", NULL);
      return;
    case Frame::EQUATORIAL_FK5:
      Tcl_AppendResult(interp, "fk5", NULL);
      return;
    case Frame::ICRS:
      Tcl_AppendResult(interp, "icrs", NULL);
      return;
    case Frame::GALACTIC:
      Tcl_AppendResult(interp, "galactic", NULL);
      return;
    case Frame::ECLIPTIC:
      Tcl_AppendResult(interp, "ecliptic", NULL);
      return;
    case Frame::LINEAR:
      Tcl_AppendResult(interp, "linear", NULL);
      return;
    case Frame::DETECTOR:
      Tcl_AppendResult(interp, "detector", NULL);
      return;
    case Frame::AMPLIFIER:
      Tcl_AppendResult(interp, "amplifier", NULL);
      return;
    case Frame::PHYSICAL:
      Tcl_AppendResult(interp, "physical", NULL);
      return;
    case Frame::IMAGE:
      Tcl_AppendResult(interp, "image", NULL);
      return;
    }
  }
  else
    Tcl_AppendResult(interp, "", NULL);
}

void Frame::getGridOptionCmd()
{
  if (grid)
    Tcl_AppendResult(interp, grid->option(), NULL);
  else
    Tcl_AppendResult(interp, "", NULL);
}

void Frame::getPixelTableCmd(const Vector& v, CoordSystem ref,
			     int size, char* var)
{
  switch (ref) {
  case CANVAS:
    getPixelTable(v, ref, &FitsImage::getCanvasToData, size, var);
    break;
  case PANNER:
    getPixelTable(v, ref, &FitsImage::getPannerToData, size, var);
    break;
  }
}

void Frame::getPixelTable(const Vector& v, CoordSystem ref,
			  Matrix& (FitsImage::*matrixToData)(),
			  int size, char* var)
{
  if (fits) {
    FitsImage* ptr = fits;
    while (ptr) {
      Vector img = v * (ptr->*matrixToData)();
      if (img[0]>=0 && img[0]<ptr->width() && 
	  img[1]>=0 && img[1]<ptr->height()) {

	Vector half((size-1)/2,(size-1)/2);
	Vector ur = img+half;
	Vector ll = img-half;
	int i,j;

	// x (columns)
	for (i=0,j=0; i<size; i++) {
	  char index[8];
	  ostrstream str(index,8);
	  str << i+1 << j << ends;

	  if (ur[0]>=0 && ll[0]<ptr->width() && 
	      ur[1]>=0 && ll[1]<ptr->height()) {

	    Vector pt = ((ll+Vector(i,j)) * Translate(.5,.5)).round();
	    if (pt[0]>0 && pt[0]<=ptr->width()) {
	      char lbuf[8];
	      ostrstream lstr(lbuf,8);
	      lstr << pt[0] << ends;
	      Tcl_SetVar2(interp,var,index,lbuf,0);
	    }
	    else
	      Tcl_SetVar2(interp,var,index,"",0);
	  }
	  else
	    Tcl_SetVar2(interp,var,index,"",0);
	}

	// y (rows)
	for (i=0,j=0; j<size; j++) {
	  char index[8];
	  ostrstream str(index,8);
	  str << i << j+1 << ends;

	  if (ur[0]>=0 && ll[0]<ptr->width() && 
	      ur[1]>=0 && ll[1]<ptr->height()) {

	    Vector pt = ((ll+Vector(i,j)) * Translate(.5,.5)).round();
	    if (pt[1]>0 && pt[1]<=ptr->height()) {
	      char lbuf[8];
	      ostrstream lstr(lbuf,8);
	      lstr << pt[1] << ends;
	      Tcl_SetVar2(interp,var,index,lbuf,0);
	    }
	    else
	      Tcl_SetVar2(interp,var,index,"",0);
	  }
	  else
	    Tcl_SetVar2(interp,var,index,"",0);
	}

	// body
	for (i=0; i<size; i++)
	  for (j=0; j<size; j++) {
	    Vector pt = ll+Vector(i,j);

	    char index[8];
	    ostrstream str(index,8);
	    str << i+1 << j+1 << ends;

	    if (pt[0]>=0 && pt[0]<ptr->width() && 
		pt[1]>=0 && pt[1]<ptr->height())
	      Tcl_SetVar2(interp,var,index,(char*)ptr->getValue(pt),0);
	    else
	      Tcl_SetVar2(interp,var,index,"",0);
	  }

	return;
      }

      ptr = ptr->next();
    }
  }

  // else return blank

  for (int i=0; i<=size; i++)
    for (int j=0; j<=size; j++) {
      char index[8];
      ostrstream str(index,8);
      str << i << j << ends;

      Tcl_SetVar2(interp,var,index,"",0);
    }
}
