#include "marker.h"
#include "frame.h"
#include "util.h"

// Marker Members Public

Marker::Marker(Frame* p, const Vector& ctr, const char* clr, const char* f,
	       const char* t, unsigned short prop)
{
  parent = p;
  center = ctr;
  properties = prop;

  id = sequenceID++;
  type[0] = '\0';

  handle = NULL;
  numHandle = 0;

  colorName = dupstr(clr);
  color = parent->getColor(colorName);
  selected = 0;

  display = parent->display;
  gc = parent->markerGC;
  gcxor = parent->markerGCXOR;

  text = dupstr(t);
  font = Tk_GetFont(parent->interp, parent->tkwin, f);

  previous_ = NULL;
  next_ = NULL;
}

Marker::~Marker()
{
  if (colorName)
    delete [] colorName;

  if (text)
    delete [] text;

  if (font)
    Tk_FreeFont(font);

  if (handle)
    delete [] handle;

  doCallBacks(&deleteCB);
}

void Marker::updateCoords(Matrix& m)
{
  center*=m;
  updateBBox();
}

void Marker::draw()
{
  // set color and solid/dash line

  XSetForeground(display, gc, color);
  noDash(gc);

  // draw handles

  drawHandles();

  // draw text

  drawText();

  // draw BBox

  //  XSetForeground(display, gc, parent->getColor("white"));
  //  Vector b = allBBox.ll * parent->canvasToWidget;
  //  Vector size = allBBox.size() - Vector(1,1);
  //  XDrawRectangle(display, parent->pixmap, gc, 
  //      (int)b[0], (int)b[1], (int)size[0], (int)size[1]);
}

void Marker::move(const Vector& v)
{
  center+=v;
  updateBBox();

  doCallBacks(&moveCB);
}

void Marker::moveTo(const Vector& v)
{
  center=v;
  updateBBox();

  doCallBacks(&moveCB);
}

void Marker::ps(int mode)
{
  // set color

  psColor(mode);

  // do text

  if (text && font) {
    Tcl_DString psFont;
    Tcl_DStringInit(&psFont);
    int psSize = Tk_PostscriptFontName(font, &psFont);

    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(font, &metrics);

    char buf[512];
    ostrstream str(buf,512);
    str << '/' << Tcl_DStringValue(&psFont) 
	<< " findfont " << psSize << " scalefont setfont" << endl;
      
    Tcl_DStringFree(&psFont);

    Vector bbc = bbox.center();
    Vector t =  Vector(bbc[0], bbox.ll[1]);

    str << t.TkCanvasPs(parent->canvas) << " moveto" << endl
	<< '(' << psQuote(text) << ')' << endl
	<< "dup stringwidth pop 2 div 0 exch sub " 
	<< metrics.descent << " rmoveto show" << endl
	<< ends;

    Tcl_AppendResult(parent->interp, buf, NULL);
  }
}

void Marker::select() {
  selected = 1;
  doCallBacks(&selectCB);
}

void Marker::unselect() {
  selected = 0;
  doCallBacks(&selectCB);
}

void Marker::toggleSelect() {
  selected = !selected;
  doCallBacks(&selectCB);
}

void Marker::setColor(const char* clr)
{
  if (colorName)
    delete [] colorName;

  colorName = dupstr(clr);
  color = parent->getColor(colorName);

  doCallBacks(&colorCB);
}

void Marker::setFont(const char* f)
{
  if (font)
    Tk_FreeFont(font);

  if (f)
    font = Tk_GetFont(parent->interp, parent->tkwin, f);
  else
    font = Tk_GetFont(parent->interp, parent->tkwin, "system");

  updateBBox();
  doCallBacks(&fontCB);
}

const char* Marker::getFont()
{
  if (font)
    return Tk_NameOfFont(font);
  else
    return NULL;
}

int Marker::onHandle(const Vector& v)
{
  // return handle number
  // work last to first for annuli

  for (int i=numHandle-1; i>=0; i--) {
    BBox bb(handle[i]);
    bb.expand(2);
    if (bb.isIn(v))
      return i+1;
  }
  return 0;
}

int Marker::getProperty(unsigned short which)
{
  return (properties & which) ? 1 : 0;
}

void Marker::setText(const char* str)
{
  if (text)
    delete [] text;
  text = dupstr(str);

  updateBBox();
  doCallBacks(&textCB);
}

void Marker::setProperty(unsigned short prop, int value)
{
  if (value)
    properties |= prop;
  else
    properties &= ~prop;

  if (prop == FIXED) // bbox will change
    updateBBox();

  doCallBacks(&propertyCB);
}

int Marker::setCallBack(Callback cbl, const char* proc, const char* arg)
{
  CallBack* cb = new CallBack(parent->interp, proc, arg);

  switch (cbl) {
  case SELECTCB:
    selectCB.append(cb);
    return TCL_OK;
  case MOVECB:
    moveCB.append(cb);
    return TCL_OK;
  case EDITCB:
    editCB.append(cb);
    return TCL_OK;
  case ROTATECB:
    rotateCB.append(cb);
    return TCL_OK;
  case DELETECB:
    deleteCB.append(cb);
    return TCL_OK;
  case TEXTCB:
    textCB.append(cb);
    return TCL_OK;
  case COLORCB:
    colorCB.append(cb);
    return TCL_OK;
  case PROPERTYCB:
    propertyCB.append(cb);
    return TCL_OK;
  case FONTCB:
    fontCB.append(cb);
    return TCL_OK;
  }

  return TCL_ERROR;
}

int Marker::deleteCallBack(Callback cbl, const char* proc)
{
  List<CallBack>* list = NULL;

  switch (cbl) {
  case SELECTCB:
    list = &selectCB;
    break;
  case MOVECB:
    list = &moveCB;
    break;
  case EDITCB:
    list = &editCB;
    break;
  case ROTATECB:
    list = &rotateCB;
    break;
  case DELETECB:
    list = &deleteCB;
    break;
  case TEXTCB:
    list = &textCB;
    break;
  case COLORCB:
    list = &colorCB;
    break;
  case PROPERTYCB:
    list = &propertyCB;
    break;
  case FONTCB:
    list = &fontCB;
    break;
  }

  if (list) {
    CallBack* cb = list->head();

    while (cb) {
      if (!strcmp(cb->proc(), proc)) {
	list->extractNext(cb);
	delete cb;
	return TCL_OK;
      }
      else
	cb = cb->next();
    }
  }
  
  return TCL_ERROR;
}

int Marker::isVisible(const BBox& b)
{
  // assume visible, prove otherwise
  // all coords are in canvas coords

  BBox bb(b);
  return 
    !((allBBox.ur[0] < bb.ll[0]) ||
    (allBBox.ll[0] > bb.ur[0]) ||
    (allBBox.ur[1] < bb.ll[1]) ||
    (allBBox.ll[1] > bb.ur[1]));
}

// Protected

void Marker::doCallBacks(List<CallBack>* list)
{
  char buf[8];
  ostrstream str(buf,8);
  str << id << ends;

  if (list->head())
    do
      if ((list->current())->eval(buf))
	cerr << "Marker Internal Error: Unable to eval Callback "
	     << (list->current())->proc() << endl << "  :" 
	     << parent->interp->result << endl;

    while (list->next());
}

void Marker::psColor(int mode)
{
  char buf[256];

  ostrstream str(buf,256);
  switch ((Frame::PSColorSpace)mode) {
  case Frame::BW:
  case Frame::GRAY:
    psColorGray(colorName, str);
    str << " setgray";
    break;
  case Frame::RGB:
    psColorRGB(colorName, str);
    str << " setrgbcolor";
    break;
  case Frame::CMYK:
    psColorCMYK(colorName, str);
    str << " setcmykcolor";
    break;
  }
  str << endl << ends;

  Tcl_AppendResult(parent->interp, buf, NULL);
}

void Marker::psIncludeExclude()
{
  if (properties & INCLUDE)
    psNoDash();
  else
    psDash();
}

void Marker::psNoDash()
{
  char buf[256];
  ostrstream str(buf,256);
  str << "[] 0 setdash" << endl << ends;
  Tcl_AppendResult(parent->interp, buf, NULL);
}

void Marker::psDash()
{
  char buf[256];
  ostrstream str(buf,256);
  str << "[8 3] 0 setdash" << endl << ends;
  Tcl_AppendResult(parent->interp, buf, NULL);
}

void Marker::calcAllBBox()
{
  allBBox = bbox;

  if (text && font) {
    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(font, &metrics);
    int width = Tk_TextWidth(font, text, strlen(text));

    Vector bbc = bbox.center();
    Vector ll =  Vector(bbc[0], bbox.ll[1]) * Translate(-width/2., 0);
    Vector ur = ll * Translate(width, -(metrics.ascent+metrics.descent));

    allBBox.bound(ll);
    allBBox.bound(ur);
  }
}

void Marker::drawHandles()
{
  // handles are of length 5

  if (selected)
    for (int i=0; i<numHandle; i++) {
      Vector v = (handle[i]*parent->canvasToWidget - Vector(2,2)).round();
      XFillRectangle(display, parent->pixmap, gc, (int)v[0], (int)v[1], 5, 5);
    }
}

void Marker::drawText()
{
  if (font)
    XSetFont(display, gc, Tk_FontId(font));

  if (text && font) {
    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(font, &metrics);
    int width = Tk_TextWidth(font, text, strlen(text));

    BBox bb = bbox * parent->canvasToWidget;
    Vector bbc = bb.center();
    Vector t =  Vector(bbc[0], bb.ll[1]) * 
      Translate(-width/2., -metrics.descent);
    XDrawString(display, parent->pixmap, gc, (int)t[0], (int)t[1], 
		text, strlen(text));
  }
}

void Marker::includeExclude()
{
  if (properties & INCLUDE)
    noDash(gc);
  else
    dash(gc);
}

void Marker::dash(GC ggc)
{
  char dlist[] = {8,3};
  XSetLineAttributes(display, ggc, 1, LineOnOffDash, CapButt, JoinMiter);
  XSetDashes(display, ggc, 0, dlist, 2);
}

void Marker::noDash(GC ggc)
{
  XSetLineAttributes(display, ggc, 1, LineSolid, CapButt, JoinMiter);
}

Vector* Marker::arrow(const Vector& p, const Vector& dd)
{
  Vector d = dd;      // stupid, however, no const warnings

  const int tip = 6;  // length from end of line to tip of arrow
  const int tail = 2; // length from end of line to tails of arrow
  const int wc = 2;   // width of arrow at end of line
  const int wt = 3;   // width of arrow at tails

  Vector* a = new Vector[6];
  a[0] = Vector(0, tip);
  a[1] = Vector(-wc, 0);
  a[2] = Vector(-wt, -tail);
  a[3] = Vector(0, 0);
  a[4] = Vector(wt, -tail);
  a[5] = Vector(wc, 0);

  Matrix m = Translate(0,-tip) * 
    Scale(1.5) * 
    Rotate(-M_PI/2) * 
    Rotate(-d.angle()) * 
    Translate(p);

  for (int i=0; i<6; i++) {
    a[i] = (a[i] * m).round();
  }

  return a;
}

void Marker::renderArrow(Drawable drawable, const Vector& p, 
			 const Vector& d, GC ggc)
{
  Vector* a = arrow(p,d);
  XPoint aa[6];
  for (int i=0; i<6; i++) {
    aa[i].x = (int)a[i][0];
    aa[i].y = (int)a[i][1];
  }

  XFillPolygon(display, drawable, ggc, aa, 6, Nonconvex, CoordModeOrigin);
  delete [] a;
}

void Marker::psArrow(const Vector& p, const Vector& d)
{
  char buf[256];
  Vector* a = arrow(p,d);
  for (int i=0; i<6; i++) {
    ostrstream str(buf,256);
    if (i==0)
      str << "newpath " << endl
	  << a[i].TkCanvasPs(parent->canvas) << " moveto" << endl << ends;
    else
      str << a[i].TkCanvasPs(parent->canvas) << " lineto" << endl << ends;

    Tcl_AppendResult(parent->interp, buf, NULL);
  }

  ostrstream str(buf,256);
  str << "closepath fill" << endl << ends;
  Tcl_AppendResult(parent->interp, buf, NULL);

  delete [] a;
}

void Marker::listProperties(ostream& str)
{
  if (strcmp("green",colorName) ||
      strcmp("helvetica 10 normal",getFont()) ||
      text[0] ||
      !(properties&EDIT) ||
      !(properties&MOVE) ||
      !(properties&ROTATE) ||
      !(properties&DELETE) ||
      properties&FIXED) {

    str << " #";

    if (strcmp("green",colorName))
      str << " color=" << colorName;

    if (strcmp("helvetica 10 normal",getFont()))
      str << " font=\"" << getFont() << '"';

    if (text[0]) // only list text if there is something to list
      str << " text={" << text << '}';

    if (!(properties&EDIT))
      str << " edit=0";

    if (!(properties&MOVE))
      str << " move=0";

    if (!(properties&ROTATE))
      str << " rotate=0";

    if (!(properties&DELETE))
      str << " delete=0";

    if (properties&FIXED)
      str << " fixed=1";
  }
}

// list

void Marker::list(ostream& str, int sys, int format, const char* delim)
{
  listCoordSystem(str, sys);
  str << ';';

  if (!(properties&INCLUDE))
    str << '-';

  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::PHYSICAL:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ')';
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ')';
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	str << type << '(' << ra << ',' << dec << ')';
      }
      break;
    }
    break;
  }

  listProperties(str);
  str << delim;
}

void Marker::listCoordSystem(ostream& str, int sys)
{
  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
    str << "image";
    return;
  case Frame::PHYSICAL:
    str << "physical";
    return;
  case Frame::DETECTOR:
    str << "detector";
    return;
  case Frame::AMPLIFIER:
    str << "amplifier";
    return;
  case Frame::EQUATORIAL_FK4:
    str << "fk4";
    return;
  case Frame::EQUATORIAL_FK5:
    str << "fk5";
    return;
  case Frame::ICRS:
    str << "icrs";
    return;
  case Frame::GALACTIC:
    str << "galactic";
    return;
  case Frame::ECLIPTIC:
    str << "ecliptic";
    return;
  case Frame::LINEAR:
    str << "linear";
    return;
  }
}

void Marker::listProsCoordSystem(ostream& str, int sys)
{
  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
    str << "logical";
    return;
  case Frame::PHYSICAL:
    str << "physical";
    return;
  case Frame::DETECTOR:
    str << "detector";
    return;
  case Frame::AMPLIFIER:
    str << "amplifier";
    return;
  case Frame::EQUATORIAL_FK4:
    str << "b1950";
    return;
  case Frame::EQUATORIAL_FK5:
    str << "j2000";
    return;
  case Frame::ICRS:
    str << "icrs";
    return;
  case Frame::GALACTIC:
    str << "galactic";
    return;
  case Frame::ECLIPTIC:
    str << "ecliptic";
    return;
  case Frame::LINEAR:
    str << "linear";
    return;
  }
}

